#!/usr/bin/env python
# Copyright 2009 Daniel Casner <www.danielcasner.org>
#
# This file is part of the ChuMPD project
#
# The ChuMPD project is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The ChuMPD project is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the ChuMPD project.  If not, see <http://www.gnu.org/licenses/>.
#

import xmlsocket, mpdclient2, sys, time, os, threading

class Translator(object):
    "A server which translates between MPD and Adobe flash."

    SIMPLE_COMMANDS = ['play', 'stop', 'pause', 'next', 'previous', 'setvol']

    def __init__(self, bind_address = '', port = 6601):
        self.server = xmlsocket.socket()
        self.server.bind((bind_address, port))
        self.server.listen(5)
        self.mpd = None
        self.host = None
        self.port = None
        self.password = None

    def run(self):
        while True:
            (client, cliaddr) = self.server.accept()
            sys.stdout.write("Got new client from %s at %s\n" % (str(cliaddr), time.ctime()))
            client.settimeout(1)
            xcli = xmlsocket.XMLSocket(client)
            while xcli.is_open:
                try:
                    req = xcli.get()
                except Exception, inst:
                    if len(inst.args) and inst.args[0] != 104: # Happens when the chumby switches widgets in the middle of a write so don't bother recording it.
                        sys.stderr.write("Error parsing instruction from client: %s\n" % inst)
                    break
                if req is not None:
                    if req[0] == 'connect':
                        try:
                            self.mpdConnect(req[1], req[2])
                        except Exception, inst:
                            sys.stderr.write("Error connecting to MPD as commanded:\n  Request: %s\n  Error: %s\n" % (req, inst))
                            xcli.close()
                            break
                    elif self.mpd:
                        try:
                            getattr(self.mpd, req[0])(*req[1:])
                        except Exception, inst:
                            sys.stderr.write("Error communicating with music player daemon: %s\n\tArguments were: %s\n" % (inst, req))
                            self.mpdReconnect()
                elif self.mpd:
                    try:
                        xcli.put(self.formatUpdate())
                    except Exception, inst:
                        if len(inst.args) and  inst.args[0] != 104: # Happens when the chumby switches widgets in the middle of a write so don't bother recording it.
                            sys.stderr.write("Error writing to client: %s\n" % str(inst))
                        xcli.close()
                        break
            sys.stdout.write("Client from %s went away at %s\n" % (str(cliaddr), time.ctime()))
            try:
                mpd.close()
            except:
                pass

    def mpdConnect(self, hostarg, portarg):
        if hostarg is not None:
            self.password, self.host = mpdclient2.parse_host(hostarg)
        else:
            self.password = ''
            self.host = ''
        if portarg is not None:
            self.port = portarg
        else:
            self.port = 0
        try:
            self.mpd = mpdclient2.connect(host=self.host, port=self.port, password=self.password)
        except Exception, inst:
            sys.stderr.write("Error connecting to MPD with args (%s, %s): %s\n" % (hostarg, portarg, str(inst)))

    def mpdReconnect(self):
        try:
            self.mpd = mpdclient2.connect(host=self.host, port=self.port, password=self.password)
        except Exception, inst:
            sys.stderr.write("Error reconnecting to mpd: %s\n" % str(inst))

    def mpdClose(self):
        if self.mpd:
            self.mpd.close()
            self.mpd = None

    def formatUpdate(self):
        retryCount = 30
        while retryCount > 0:
            try:
                status = self.mpd.status()
                song   = self.mpd.currentsong()
            except Exception, inst:
                sys.stderr.write("Error communicating with MPD to get update, attempting reconnect.\n")
                time.sleep(0.5)
                retryCount -= 1
                self.mpdReconnect()
            else:
                break
        if retryCount == 0:
            raise Exception, "Cannot reconnect to MPD under current conditions."
        try:
            songTime, songLength = status.time.split(':')
            songPos = int(100.0*float(songTime)/float(songLength))
        except:
            songPos = 0
        return {
            'repeat': bool(int(status.repeat)),
            'random': bool(int(status.random)),
            'volume': int(status.volume),
            'state':  status.state,
            'time':   songPos,
            'album': unicode(song.get('album', ''), 'utf-8').encode('ascii', 'replace'),
            'title': unicode(song.get('title', ''), 'utf-8').encode('ascii', 'replace'),
            'artist': unicode(song.get('artist', ''), 'utf-8').encode('ascii', 'replace'),
        }
        

class PlayerSlave(threading.Thread):

    EVENT_FILE = '/tmp/flashplayer.event'
    UPDATE_CMD = 'chumbyflashplayer.x -F1'
    PLAY_EVENT = '<event type="UserPlayer" value="play" comment="%s"/>'
    STOP_EVENT = '<event type="MusicPlayer" value="stop" comment=""/>'

    def __init__(self, music_root='/mnt/music', mpd_host=None, mpd_port=None, mpd_password=None):
        self.mpd = mpdclient2.connect(host=mpd_host, port=mpd_port, password=mpd_password)
        self.root = music_root
        threading.Thread.__init__(self)

    def run(self):
        self.stopped = False
        last_file = None
        while not self.stopped:
            cur_file = self.mpd.currentsong().get('file', None)
            if cur_file != last_file:
                if cur_file is not None: self.play_music(cur_file)
                else: self.stop_music()
                last_file = cur_file
            time.sleep(0.01)

    def stop(self):
        self.stopped = True

    def issue_event(self, event):
        try:
            evf = file(self.EVENT_FILE, 'w')
            evf.write(event)
            evf.close()
        except Exception, inst:
            sys.stdwrr.write('Error writing event to control panel: %s' % str(inst))
        else:
            os.system(self.UPDATE_CMD)
        

    def play_music(self, music):
        self.issue_event(self.PLAY_EVENT % os.path.join(self.root, music))

    def stop_music(self):
        self.issue_event(self.STOP_EVENT)
    
if __name__ == '__main__':
    nargs = len(sys.argv)
    if nargs > 1 and sys.argv[1] == 'slave':
        root = sys.argv[2] if nargs > 2 else '/mnt/music'
        host = sys.argv[3] if nargs > 3 else None
        port = sys.argv[4] if nargs > 4 else None
        pswd = sys.argv[5] if nargs > 5 else None
        s = PlayerSlave(root, host, port, pswd)
        sys.stdout.write("Starting player slave.\n")
        s.start()
        try:
            while s.isAlive():
                time.sleep(61)
        except KeyboardInterrupt:
            s.stop()
    else:
        sys.stdout.write("Starting Flash XMLSocket MPD translation server.\n")
        t = Translator()
        try:
            t.run()
        except KeyboardInterrupt:
            sys.stdout.write("Translator closed by user.\n")
            t.server.close()
