#!/usr/bin/env python
# Copyright 2009 Daniel Casner <www.danielcasner.org>
#
# This file is part of the ChuMPD project
#
# The ChuMPD project is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# The ChuMPD project is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with the ChuMPD project.  If not, see <http://www.gnu.org/licenses/>.
#

from xml.dom import minidom
from socket import *

ENCODING='utf-8'

# XML representation objects and methods: ######################################

BASIC_TYPES_2XML = {
    bool:  u'bool',
    int:   u'int',
    float: u'float',
    str:   u'string'
}

BASIC_TYPES_2PY = {
    u'bool':   lambda x: x == 'True',
    u'int':    int,
    u'float':  float,
    u'string': str
}


class XMLRep(minidom.Document):
    "An XML representation of a python structure consisting of supported types."
    def __init__(self, *args):
        "Initalizes the XML representation, optionally with a python structure."
        minidom.Document.__init__(self)
        if len(args) > 1: raise TypeError, "XMLRep takes exactly 0 or 1 arguments, given " + str(len(args))
        if len(args) > 0:
            self.appendChild(self.xmlify(args[0]))
    
    def basicType(self, v):
        "Converts a basic python type to an XML node representation of it."
        node = self.createElement(BASIC_TYPES_2XML[type(v)])
        node.appendChild(self.createTextNode(str(v).encode(ENCODING, 'replace')))
        return node
    
    def dict(self, d):
        "Converts a python dict instance to an XML node tree representation of it."
        node = self.createElement('dict')
        for key, value in d.iteritems():
            item = self.createElement('item')
            ik = self.createElement('key')
            ik.appendChild(self.xmlify(key))
            iv = self.createElement('value')
            iv.appendChild(self.xmlify(value))
            item.appendChild(ik)
            item.appendChild(iv)
            node.appendChild(item)
        return node

    def list(self, a):
        "Converts a python list or tuple instance to an XML node tree representation of it."
        if isinstance(a, list): node = self.createElement('list')
        elif isinstance(a, tuple): node = self.createElement('tuple')
        else: raise TypeError, "Unsupported type past to XMLResponse.list: " + str(type(a))
        for i in a:
            node.appendChild(self.xmlify(i))
        return node

    def none(self):
        "Converts None to an XML node representation of it."
        return self.createElement('None')

    def xmlify(self, data):
        "Converts a python structure consisting of an arbitrary tree of supported types into an XML node tree representation of it."
        if isinstance(data, list) or isinstance(data, tuple):
            return self.list(data)
        elif isinstance(data, dict):
            return self.dict(data)
        elif data is None:
            return self.none()
        elif type(data) in BASIC_TYPES_2XML.keys():
            return self.basicType(data)
        else:
            raise TypeError, "XMLResponse, unsupported type conversion: " + str(type(data))


def pyify_dict(node):
    "Converts an XML representation of a python dict back into a python object."
    ret = {}
    for item in node.childNodes:
        if item.nodeName != u'item':
            raise ValueError, 'dict tag contained non "item" tag: ' + item.nodeName
        else:
            xk, xv = item.childNodes
            if xk.nodeName != u'key':
                raise ValueError, 'Expected "key" tag, got: ' + xk.nodeName
            elif xv.nodeName != u'value':
                raise ValueError, 'Expected "value" tag, got: ' + xv.nodeName
            else:
                ret[pyify(xk.firstChild)] = pyify(xv.firstChild)
    return ret

def pyify_list(node):
    "Converts an XML representation of a python list or tuple back into a python object."
    ret = [pyify(child) for child in node.childNodes]
    if node.nodeName == u'tuple':
        return tuple(ret)
    else:
        return ret

def pyify_basic(node):
    "Converts an XML representation of a python basic type back into a python object."
    if node.nodeName == u'string' and node.firstChild is None: return ''
    else: return BASIC_TYPES_2PY[node.nodeName](node.firstChild.data)

def pyify(node):
    "Converts an XML node tree generated by XMLRep back into a python structure."
    if node.nodeName == u'dict':
        return pyify_dict(node)
    elif node.nodeName == u'list' or node.nodeName == u'tuple':
        return pyify_list(node)
    elif node.nodeName == u'None':
        return None
    elif node.nodeName in BASIC_TYPES_2PY.keys():
        return pyify_basic(node)
    else:
        raise TypeError, "pyify, unsupported XML tag: " + node.nodeName


# Socket objects and methods: ##################################################

class XMLSocket(object):
    'The other end of an Adobe flash 8 XMLSocket. Use the "get" and "put" methods instances of this class to communicate with your flash programs.'
    def __init__(self, *args, **kwargs):
        if len(args) and isinstance(args[0], socket):
            self.sock = args[0]
        else:
            self.sock = socket(*args, **kwargs)
        self.xml_data = ''
        self.is_open = True

    def connect(self, addr):
        return self.sock.connect(addr)
    
    def close(self):
        self.is_open = False
        self.sock.close()

    def get(self):
        "Receive an xml encoded structure from the socket."
        eop = self.xml_data.find('\0')
        while eop == -1:
            try:
                d = self.sock.recv(100)
                if d == '':
                    self.is_open = False
                    return None
                else:
                    self.xml_data += d
            except timeout:
                return None
            eop = self.xml_data.find('\0')
        pkt = self.xml_data[:eop]
        self.xml_data = self.xml_data[eop+1:]
        x = minidom.parseString(pkt)
        if x.firstChild.nodeName == 'policy-file-request':
            self.sock.send(file('crossdomain.xml', 'r').read() + '\0')
        else:
            return pyify(x.firstChild)

    def put(self, struct):
        pkt = XMLRep(struct).toxml(ENCODING) + '\0'
        gone = self.sock.send(pkt)
        while gone < len(pkt):
            pkt = pkt[gone:]
            gone = self.sock.send(pkt)
        
